package com.bureau.devicefingerprint

import android.Manifest
import android.content.Context
import android.content.pm.PackageManager
import android.util.Log
import androidx.annotation.Keep
import androidx.core.app.ActivityCompat
import com.bureau.base.Environment
import com.bureau.devicefingerprint.data.remote.RemoteDataSource
import com.bureau.devicefingerprint.data.remote.RemoteDataSourceImpl
import com.bureau.devicefingerprint.models.Callback
import com.bureau.devicefingerprint.models.CredentialBuilder
import com.bureau.devicefingerprint.models.ErrorResponse
import com.bureau.devicefingerprint.models.SubmitResponse
import com.bureau.devicefingerprint.models.devicedataholder.DeviceDataHolder
import com.bureau.devicefingerprint.models.devicedataholder.RealTimeDatHolder
import com.bureau.devicefingerprint.tools.DataCallback
import com.bureau.devicefingerprint.tools.RealTimeDataCallback
import com.bureau.devicefingerprint.tools.RealTimeDataChannel
import com.bureau.devicefingerprint.tools.SharedPreferenceHelper
import com.bureau.devicefingerprint.tools.SystemPropertiesInvoker
import com.bureau.onetaplogin.BureauAuth
import com.bureau.onetaplogin.models.AuthCallback
import com.bureau.onetaplogin.models.AuthenticationStatus
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch

@Keep
class BureauAPI private constructor(private val context: Context, clientId: String, private val environment: Environment, private val isAuthEnabled: Boolean) {
    private val remoteDataSource: RemoteDataSource by lazy { RemoteDataSourceImpl(environment) }
    private var currentParentSessionId: String = ""
    private var isStartSessionCalled = false
    private var credentialBuilder: CredentialBuilder? = CredentialBuilder()
    private var deviceData: DeviceDataHolder? = null
    private var dataMapper: DataMapper = DataMapper(context)
    private var dataMonitoring: DataMonitoring = DataMonitoring(context)
    private var permissionList: ArrayList<String> = arrayListOf(Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION)
    private var isRealTimeDataEnabled: Boolean = false
    private var realTimeDataHolder: RealTimeDatHolder? = null
    private var realTimeDataCallback: RealTimeDataCallback? = null

    private val bureauAuth: BureauAuth by lazy {
        BureauAuth.Builder()
            .environment(environment)
            .clientId(clientId)
            .build()
    }

    init {
        if (clientId.isEmpty()) {
            throw Error("Credential id should not be empty")
        }
        credentialBuilder?.merchantId = clientId
    }

    fun setSessionId(sessionId: String) {
        credentialBuilder?.sessionId = sessionId
        currentParentSessionId = sessionId
    }

    fun startSession(authCallback: AuthCallback) {
        if (isAuthEnabled) {
            registerAuth(authCallback)
        }
        checkErrors()
        collectData()
        isStartSessionCalled = true
    }

    fun setUserId(userId: String) {
        credentialBuilder?.userId = userId
    }

    private fun registerAuth(authCallback: AuthCallback) {
        CoroutineScope(Dispatchers.IO).launch {
            if (checkOTLErrors()) {
                setBureauAuth(bureauAuth, context, authCallback)
            } else {
                authCallback.onResult(AuthenticationStatus.UnknownState)
            }
        }
    }

    fun setFlow(flow: String) {
        credentialBuilder?.flow = flow
    }

    private fun collectData() {
        var hasLocationPermission = false
        if (hasPermissions(context)) {
            hasLocationPermission = true
        }
        deviceData = dataMapper.getSynchronisedOSData(hasLocationPermission)
    }

    fun setRealtimeCallback(callback: RealTimeDataCallback) {
        this.realTimeDataCallback = callback
        isRealTimeDataEnabled = true
        realTimeDataHolder = RealTimeDatHolder()
        Log.i("DeviceService", "Real Time Data Monitoring Started")
        dataMonitoring.registerDataMonitoring(object : RealTimeDataChannel {
            override fun emitVPNStatus(item: Boolean) {
                realTimeDataHolder?.vpnStatus_ = item
                if (item) {
                    SharedPreferenceHelper.setSharedPreferenceBoolean(
                        context,
                        SharedPreferenceHelper.isPastVPNData,
                        value = true,
                        isDataMonitoring = true
                    )
                }
                CoroutineScope(Dispatchers.Main).launch {
                    realTimeDataCallback?.onVPNStatusChange(item)
                }
            }

            override fun emitMockGPSStatus(item: Boolean) {
                realTimeDataHolder?.mockGPSStatus_ = item
                if (item) {
                    SharedPreferenceHelper.setSharedPreferenceBoolean(
                        context,
                        SharedPreferenceHelper.isPastMockGPSData,
                        value = true,
                        isDataMonitoring = true
                    )
                }
                CoroutineScope(Dispatchers.Main).launch {
                    realTimeDataCallback?.onMockGpsStatusChange(item)
                }
            }

            override fun emitDebuggerStatus(item: Boolean) {
                realTimeDataHolder?.debuggerStatus_ = item
                if (item) {
                    SharedPreferenceHelper.setSharedPreferenceBoolean(
                        context,
                        SharedPreferenceHelper.isPastDebuggerData,
                        value = true,
                        isDataMonitoring = true
                    )
                }
                CoroutineScope(Dispatchers.Main).launch {
                    realTimeDataCallback?.onDebuggerStatusChange(item)
                }
            }
        })
    }

    fun submit(callback: DataCallback) {
        submitData(callback, currentParentSessionId)
    }

    private fun submitData(callback: DataCallback, sessionId: String) {
        if (!isStartSessionCalled) {
            setUpMandatoryDataCollection()
        }

        CoroutineScope(Dispatchers.IO).launch {
            deviceData =
                deviceData?.let {
                    dataMapper.getAsyncOSData(
                        credentialBuilder = credentialBuilder,
                        it
                    )
                }
            deviceData = deviceData?.let { dataMapper.getAsynchronousIds(it) }
            deviceData = deviceData?.let { dataMapper.addGeocodeData(it) }
            if (isRealTimeDataEnabled) {
                deviceData?.realTimeDataInfo_ = realTimeDataHolder
            }

            sendData(sessionId, callback, "")
        }
    }

    private suspend fun sendData(
        sessionId: String,
        callback: DataCallback,
        networkSessionId: String
    ) {
        deviceData?.let { dataMapper.addPastRealData(it) }

        if (sessionId.isNotEmpty()) {
            deviceData?.sessionId_ = sessionId
        }
        if (networkSessionId.isNotEmpty()) {
            deviceData?.networkSessionId_ = networkSessionId
        }

        remoteDataSource.sendData(
            deviceData = deviceData,
            object : Callback {
                override fun getResult(isSuccess: Boolean?) {
                    SharedPreferenceHelper.clearAllDataPref(context)

                    CoroutineScope(Dispatchers.Main).launch {
                        isStartSessionCalled = false
                        callback.onResult(SubmitResponse("Success"))
                    }
                }

                override fun getError(error: String?, errorCode: String) {
                    SharedPreferenceHelper.clearAllDataPref(context)
                    CoroutineScope(Dispatchers.Main).launch {
                        isStartSessionCalled = false

                        callback.onError(ErrorResponse(error ?: "Error", errorCode))
                    }
                }
            },
            credentialBuilder?.merchantId ?: ""
        )
    }

    private fun hasPermissions(context: Context): Boolean {
        permissionList.forEach {
            if (ActivityCompat.checkSelfPermission(
                    context,
                    it
                ) != PackageManager.PERMISSION_GRANTED
            ) {
                return false
            }
        }
        return true
    }

    private fun checkErrors() {
        if (credentialBuilder?.sessionId?.isEmpty() == true) {
            throw Error("Session id should not be empty")
        }
        if (credentialBuilder?.userId?.isEmpty() == true) {
            Log.w("Prism SDK", "User Id received is empty")
        }
    }

    private fun checkOTLErrors(): Boolean {
        if (credentialBuilder?.userId?.isEmpty() == true) {
            Log.w(
                "Prism SDK",
                "User ID is empty.OTL cannot bisDebuggerFirede triggered. Please call setUserID method record a userID"
            )
            return false
        }

        if (!SystemPropertiesInvoker.isUserValidForOTL(credentialBuilder?.userId ?: "")) {
            Log.w("Prism SDK", "User Id is not a number. Terminating UserAuth Process")
            return false
        }
        return true
    }

    private fun setBureauAuth(bureauAuth: BureauAuth, context: Context, authCallback: AuthCallback) {
        if (SystemPropertiesInvoker.isUserValidForOTL(credentialBuilder?.userId ?: "")) {
            bureauAuth.authenticate(
                context,
                credentialBuilder?.sessionId ?: "",
                credentialBuilder?.userId?.toLong() ?: 0,
                authCallback
            )
        }
    }

    private fun setUpMandatoryDataCollection() {
        checkErrors()
        collectData()
    }

    fun removeRealTimeCallBack() {
        if (isRealTimeDataEnabled) {
            this.realTimeDataCallback = null
            isRealTimeDataEnabled = false
            dataMonitoring.unRegisterDataMonitoring()
            Log.i("DeviceService", "Real Time Data Monitoring Stopped")
        }
    }

    @Keep
    companion object {
        fun getBureauApiInstance(context: Context, clientId: String, environment: Environment = Environment.ENV_PRODUCTION, enableAuth: Boolean = false): BureauAPI {
            return BureauAPI(context.applicationContext, clientId, environment, enableAuth)
        }
    }
}
